local ADDON, Package = ...

local Module = {}
local Registry = {}

local Preserve_CraftFrame_Update = nil
local Preserve_TradeSkillFrame_Update = nil

local clone = Package.Enum.clone
local create_searchbox = Package.Widget.create_searchbox
local create_checkbox = Package.Widget.create_checkbox
local purge_keyboard_focus = Package.Widget.purge_keyboard_focus

local agent = CreateFrame("Frame")

local initial_state = {
  trade_search_term = "",
  trade_searched_recipes = {},
  craft_search_term = "",
  craft_searched_recipes = {}
}

local state = clone(initial_state)

function Module.rollback_craft()
  state.craft_search_term = ""
  state.craft_searched_recipes = {}
  CraftFrame.SearchBox:SetText("")
  CraftFrame_Update = Preserve_CraftFrame_Update
end

function Module.rollback_trade()
  state.trade_search_term = ""
  state.trade_searched_recipes = {}
  TradeSkillFrame.SearchBox:SetText("")
  TradeSkillFrame_Update = Preserve_TradeSkillFrame_Update
end

function Module.search_craft_recipes()
  state.craft_searched_recipes = {}

  local craft_length = GetNumCrafts()

  for craft_index = 1, craft_length do
    local craft_name, _, craft_type, craft_available = GetCraftInfo(craft_index)
    local craft_desc = GetCraftDescription(craft_index)
    local reagent_length = GetCraftNumReagents(craft_index)
    local reagent_names = ""

    for reagent_index = 1, reagent_length do
      local reagent_name = GetCraftReagentInfo(craft_index, reagent_index)

      if reagent_name then
        reagent_names = reagent_names .. " " .. reagent_name:gsub(" ", "")
      end
    end

    if (CraftFrame.CheckBox:GetChecked() and craft_available > 0 and craft_type ~= "header")
    or (not CraftFrame.CheckBox:GetChecked() and state.craft_search_term ~= "" and craft_type ~= "header") then
      local find_by_name = craft_name and string.find(craft_name:gsub(" ", ""), state.craft_search_term, 1, true)
      local find_by_desc = craft_desc and string.find(craft_desc:gsub(" ", ""), state.craft_search_term, 1, true)
      local find_by_reagent = reagent_names:len() > 0 and string.find(reagent_names, state.craft_search_term, 1, true)

      if find_by_name or find_by_desc or find_by_reagent then
        tinsert(state.craft_searched_recipes, craft_index)
      end
    end
  end
end

function Module.update_craft_recipes(options)
  if state.craft_search_term ~= "" or CraftFrame.CheckBox:GetChecked() then
    if options.scroll_to_top then
      CraftListScrollFrameScrollBar:SetValue(0)
    end

    if options.select_first_item and #state.craft_searched_recipes > 0 then
      CraftFrame_SetSelection(state.craft_searched_recipes[1])
    end

    CraftFrame_Update = Override_CraftFrame_Update
    CraftFrame_Update()

  else
    if not options.prevent_default then
      CraftFrame_Update = Preserve_CraftFrame_Update
      CraftFrame_OnShow(CraftFrame)
    end
  end
end

function Module.search_trade_recipes()
  state.trade_searched_recipes = {}

  local skill_length = GetNumTradeSkills()

  for skill_index = 1, skill_length do
    local skill_name_raw, skill_type, skill_available = GetTradeSkillInfo(skill_index)
    local skill_name = skill_name_raw:gsub(" ", "")
    local reagent_length = GetTradeSkillNumReagents(skill_index)
    local reagent_names = ""

    for reagent_index = 1, reagent_length do
      local reagent_name = GetTradeSkillReagentInfo(skill_index, reagent_index)

      if reagent_name then
        reagent_names = reagent_names .. " " .. reagent_name:gsub(" ", "")
      end
    end

    if (TradeSkillFrame.CheckBox:GetChecked() and skill_available > 0 and skill_type ~= "header")
    or (not TradeSkillFrame.CheckBox:GetChecked() and state.trade_search_term ~= "" and skill_type ~= "header") then
      local find_by_skill_name = string.find(skill_name, state.trade_search_term, 1, true)
      local find_by_reagent_name = string.find(reagent_names, state.trade_search_term, 1, true)

      if find_by_skill_name or find_by_reagent_name then
        tinsert(state.trade_searched_recipes, skill_index)
      end
    end
  end
end

function Module.update_trade_recipes(options)
  if state.trade_search_term ~= "" or TradeSkillFrame.CheckBox:GetChecked() then
    if options.scroll_to_top then
      TradeSkillListScrollFrameScrollBar:SetValue(0)
    end

    if options.select_first_item and #state.trade_searched_recipes > 0 then
      TradeSkillFrame_SetSelection(state.trade_searched_recipes[1])
    end

    TradeSkillFrame_Update = Overidde_TradeSkillFrame_Update
    TradeSkillFrame_Update()
  else
    if not options.prevent_default then
      TradeSkillFrame_Update = Preserve_TradeSkillFrame_Update
      TradeSkillFrame_SetSelection(0)
      TradeSkillFrame_OnShow(TradeSkillFrame)
    end
  end
end

function Module.on_craft_searchbox_text_changed(self)
  local text = self:GetText():trim():gsub(" ", "")

  if state.craft_search_term ~= text then
    state.craft_search_term = text

    Module.search_craft_recipes()
    Module.update_craft_recipes({scroll_to_top = true, select_first_item = true})
  end
end

function Module.on_craft_checkbox_click(self)
  Module.search_craft_recipes()
  Module.update_craft_recipes({scroll_to_top = true, select_first_item = true})
end

function Module.on_trade_searchbox_text_changed(self)
  local text = self:GetText():trim():gsub(" ", "")

  if state.trade_search_term ~= text then
    state.trade_search_term = text

    Module.search_trade_recipes()
    Module.update_trade_recipes({scroll_to_top = true, select_first_item = true})
  end
end

function Module.on_trade_checkbox_click(self)
  Module.search_trade_recipes()
  Module.update_trade_recipes({scroll_to_top = true, select_first_item = true})
end

function Module.enable()
  agent:RegisterEvent("CRAFT_SHOW")
  agent:RegisterEvent("CRAFT_CLOSE")
  agent:RegisterEvent("CRAFT_UPDATE")
  agent:RegisterEvent("TRADE_SKILL_SHOW")
  agent:RegisterEvent("TRADE_SKILL_CLOSE")
  agent:RegisterEvent("TRADE_SKILL_UPDATE")
  agent:SetScript("OnEvent", Module.on_event)
end

function Module.disable()
  agent:UnregisterEvent("CRAFT_SHOW")
  agent:UnregisterEvent("CRAFT_CLOSE")
  agent:UnregisterEvent("CRAFT_UPDATE")
  agent:UnregisterEvent("TRADE_SKILL_SHOW")
  agent:UnregisterEvent("TRADE_SKILL_CLOSE")
  agent:UnregisterEvent("TRADE_SKILL_UPDATE")
  agent:SetScript("OnEvent", nil)
end

function Module.on_event(self, event, ...)
  Registry[event](self, ...)
end

function Registry.CRAFT_SHOW()
  if (not CraftFrame.loaded) then
    local frame_width = CraftRankFrame:GetWidth()
    local frame_skill_rank_width = CraftRankFrameSkillRank:GetWidth()
    local frame_border_width = CraftRankFrameBorder:GetWidth()

    CraftRankFrame:SetWidth(frame_width / 2)
    CraftRankFrameBorder:SetWidth(frame_border_width / 2)
    CraftRankFrameSkillRank:SetWidth(frame_skill_rank_width / 2)

    local frame = create_searchbox({}, CraftFrame)
    frame:ClearAllPoints()
    frame:SetSize(110, 18)
    frame:SetPoint("LEFT", CraftRankFrameBorder, "RIGHT", 5, 0)
    frame:HookScript("OnTextChanged", Module.on_craft_searchbox_text_changed)

    local checkbox = create_checkbox(nil, CraftFrame)
    checkbox:SetSize(20, 20)
    checkbox:SetPoint("LEFT", frame, "RIGHT", -1, -1)
    checkbox:SetScript("OnClick", Module.on_craft_checkbox_click)
    checkbox.text:Hide()

    CraftFrame.loaded = true
    CraftFrame.CheckBox = checkbox
    CraftFrame.SearchBox = frame
    CraftFrame:HookScript("OnMouseDown", purge_keyboard_focus)

    Preserve_CraftFrame_Update = CraftFrame_Update
  end

  Module.rollback_craft()
end

function Registry.CRAFT_CLOSE()
  Module.rollback_craft()
end

function Registry.CRAFT_UPDATE()
  Module.search_craft_recipes()
  Module.update_craft_recipes({prevent_default = true})
end

function Registry.TRADE_SKILL_SHOW()
  if (not TradeSkillFrame.loaded) then
    local frame_width = TradeSkillRankFrame:GetWidth()
    local frame_skill_rank_width = TradeSkillRankFrameSkillRank:GetWidth()
    local frame_border_width = TradeSkillRankFrameBorder:GetWidth()

    TradeSkillRankFrame:SetWidth(frame_width / 2)
    TradeSkillRankFrameBorder:SetWidth(frame_border_width / 2)
    TradeSkillRankFrameSkillRank:SetWidth(frame_skill_rank_width / 2)

    local frame = create_searchbox({}, TradeSkillFrame)
    frame:ClearAllPoints()
    frame:SetSize(110, 18)
    frame:SetPoint("LEFT", TradeSkillRankFrameBorder, "RIGHT", 5, 0)
    frame:HookScript("OnTextChanged", Module.on_trade_searchbox_text_changed)

    local checkbox = create_checkbox(nil, TradeSkillFrame)
    checkbox:SetSize(20, 20)
    checkbox:SetPoint("LEFT", frame, "RIGHT", -1, -1)
    checkbox:SetScript("OnClick", Module.on_trade_checkbox_click)
    checkbox.text:Hide()

    TradeSkillFrame.loaded = true
    TradeSkillFrame.CheckBox = checkbox
    TradeSkillFrame.SearchBox = frame
    TradeSkillFrame:HookScript("OnMouseDown", purge_keyboard_focus)

    Preserve_TradeSkillFrame_Update = TradeSkillFrame_Update
  end

  -- TODO: WOW BUG: Blizzard_TradeSkillUI DropDown keeps selecting wrong items.
  -- UIDropDownMenu_SetSelectedID(TradeSkillSubClassDropDown, 1)
  -- SetTradeSkillSubClassFilter(0, 1, 1)
  -- UIDropDownMenu_SetSelectedID(TradeSkillInvSlotDropDown, 1)
  -- SetTradeSkillInvSlotFilter(0, 1, 1)

  Module.rollback_trade()
end

function Registry.TRADE_SKILL_CLOSE()
  Module.rollback_trade()
end

function Registry.TRADE_SKILL_UPDATE()
  Module.search_trade_recipes()
  Module.update_trade_recipes({prevent_default = true})
end

-- BlizzardInterfaceCode\Interface\AddOns\Blizzard_CraftUI\Blizzard_CraftUI.lua
function Override_CraftFrame_Update()
  SetPortraitTexture(CraftFramePortrait, "player")
  CraftFrameTitleText:SetText(GetCraftName())

  local numCrafts = #state.craft_searched_recipes
  local craftOffset = FauxScrollFrame_GetOffset(CraftListScrollFrame)
  CraftCreateButton:SetText(getglobal(GetCraftButtonToken()))
  local name, rank, maxRank = GetCraftDisplaySkillLine()
  if ( name ) then
    CraftRankFrameSkillName:SetText(name)
    CraftRankFrame:SetStatusBarColor(0.0, 0.0, 1.0, 0.5)
    CraftRankFrameBackground:SetVertexColor(0.0, 0.0, 0.75, 0.5)
    CraftRankFrame:SetMinMaxValues(0, maxRank)
    CraftRankFrame:SetValue(rank)
    CraftRankFrameSkillRank:SetText(rank.."/"..maxRank)
    CraftRankFrame:Show()
    CraftSkillBorderLeft:Show()
    CraftSkillBorderRight:Show()
  else
    CraftRankFrame:Hide()
    CraftSkillBorderLeft:Hide()
    CraftSkillBorderRight:Hide()
  end

  if ( numCrafts <=1 ) then
    CraftExpandButtonFrame:Hide()
  else
    CraftExpandButtonFrame:Show()
  end

  if ( numCrafts == 0 ) then
    CraftName:Hide()
    CraftRequirements:Hide()
    CraftIcon:Hide()
    CraftReagentLabel:Hide()
    CraftDescription:Hide()
    for i=1, MAX_CRAFT_REAGENTS, 1 do
      getglobal("CraftReagent"..i):Hide()
    end
    CraftDetailScrollFrameScrollBar:Hide()
    CraftDetailScrollFrameTop:Hide()
    CraftDetailScrollFrameBottom:Hide()
    CraftListScrollFrame:Hide()
    for i=1, CRAFTS_DISPLAYED, 1 do
      getglobal("Craft"..i):Hide()
    end
    CraftHighlightFrame:Hide()
    CraftRequirements:Hide()
    return
  end

  CraftName:Show()
  CraftRequirements:Show()
  CraftIcon:Show()
  CraftDescription:Show()
  CraftCollapseAllButton:Enable()

  FauxScrollFrame_Update(CraftListScrollFrame, numCrafts, CRAFTS_DISPLAYED, CRAFT_SKILL_HEIGHT, nil, nil, nil, CraftHighlightFrame, 293, 316 )

  CraftHighlightFrame:Hide()

  local craftIndex, craftName, craftButton, craftButtonSubText, craftButtonCost, craftButtonText
  for i=1, CRAFTS_DISPLAYED, 1 do
    craftAt = i + craftOffset
    craftIndex = state.craft_searched_recipes[craftAt] or 0
    craftName, craftSubSpellName, craftType, numAvailable, isExpanded, trainingPointCost, requiredLevel = GetCraftInfo(craftIndex)
    craftButton = getglobal("Craft"..i)
    craftButtonSubText = getglobal("Craft"..i.."SubText")
    craftButtonCost = getglobal("Craft"..i.."Cost")
    craftButtonText = getglobal("Craft"..i.."Text")
    if ( craftName and craftAt <= numCrafts ) then
      if ( CraftListScrollFrame:IsVisible() ) then
        craftButton:SetWidth(293)

        if ( not trainingPointCost ) then
          craftButtonText:SetWidth(290)
        end
      else
        craftButton:SetWidth(323)

        if ( not trainingPointCost ) then
          craftButtonText:SetWidth(320)
        end
      end
      craftButtonCost:SetPoint("RIGHT", CRAFT_COST_OFFSET, 0)
      local color = CraftTypeColor[craftType]

      craftButton:SetNormalFontObject(color.font)
      craftButtonCost:SetTextColor(color.r, color.g, color.b)
      Craft_SetSubTextColor(craftButton, color.r, color.g, color.b)
      craftButton:SetID(craftIndex)
      craftButton:Show()

      if ( craftType == "header" ) then
        craftButton:SetText(craftName)
        craftButtonSubText:SetText("")
        craftButtonText:SetWidth(0)
        if ( isExpanded ) then
          craftButton:SetNormalTexture("Interface\\Buttons\\UI-MinusButton-Up")
        else
          craftButton:SetNormalTexture("Interface\\Buttons\\UI-PlusButton-Up")
        end
        getglobal("Craft"..i.."Highlight"):SetTexture("Interface\\Buttons\\UI-PlusButton-Hilight")
        getglobal("Craft"..i):UnlockHighlight()
      else
        craftButton:SetNormalTexture("")
        getglobal("Craft"..i.."Highlight"):SetTexture("")
        if ( numAvailable == 0 ) then
          craftButton:SetText(" "..craftName)
        else
          craftButton:SetText(" "..craftName.." ["..numAvailable.."]")
        end
        if ( craftSubSpellName and craftSubSpellName ~= "" ) then
          craftButtonSubText:SetText(format(PARENS_TEMPLATE, craftSubSpellName))
          craftButtonText:SetWidth(0)
        else
          craftButtonSubText:SetText("")

          craftButtonText:SetWidth(CRAFT_TEXT_WIDTH)
        end
        if ( trainingPointCost > 0 ) then
          craftButtonCost:SetText(format(TRAINER_LIST_TP, trainingPointCost))
        else
          craftButtonCost:SetText("")
        end

        craftButtonSubText:SetPoint("LEFT", "Craft"..i.."Text", "RIGHT", CRAFT_SUBTEXT_OFFSET, 0)

        if ( GetCraftSelectionIndex() == craftIndex ) then
          CraftHighlightFrame:SetPoint("TOPLEFT", "Craft"..i, "TOPLEFT", 0, 0)
          CraftHighlightFrame:Show()
          Craft_SetSubTextColor(craftButton, 1.0, 1.0, 1.0)
          craftButtonCost:SetTextColor(1.0, 1.0, 1.0)
          getglobal("Craft"..i):LockHighlight()
        else
          getglobal("Craft"..i):UnlockHighlight()
        end
      end

    else
      craftButton:Hide()
    end
  end

  Craft_UpdateTrainingPoints()

  local numHeaders = 0
  local notExpanded = 0
  for i=1, numCrafts, 1 do
    local craftName, craftSubSpellName, craftType, numAvailable, isExpanded = GetCraftInfo(i)
    if ( craftName and craftType == "header" ) then
      numHeaders = numHeaders + 1
      if ( not isExpanded ) then
        notExpanded = notExpanded + 1
      end
    end
  end

  if ( notExpanded ~= numHeaders ) then
    CraftCollapseAllButton.collapsed = nil
    CraftCollapseAllButton:SetNormalTexture("Interface\\Buttons\\UI-MinusButton-Up")
  else
    CraftCollapseAllButton.collapsed = 1
    CraftCollapseAllButton:SetNormalTexture("Interface\\Buttons\\UI-PlusButton-Up")
  end

  if ( numHeaders > 0 ) then
    for i=1, CRAFTS_DISPLAYED, 1 do
      getglobal("Craft"..i.."Text"):SetPoint("TOPLEFT", "Craft"..i, "TOPLEFT", 21, 0)
    end
    CraftExpandButtonFrame:Show()
  else
    for i=1, CRAFTS_DISPLAYED, 1 do
      getglobal("Craft"..i.."Text"):SetPoint("TOPLEFT", "Craft"..i, "TOPLEFT", 3, 0)
    end
    CraftExpandButtonFrame:Hide()
  end
end

-- BlizzardInterfaceCode\Interface\AddOns\Blizzard_TradeSkillUI\Blizzard_TradeSkillUI.lua
function Overidde_TradeSkillFrame_Update()
  local numTradeSkills = #state.trade_searched_recipes
  local skillOffset = FauxScrollFrame_GetOffset(TradeSkillListScrollFrame)
  if ( numTradeSkills == 0 ) then
    TradeSkillFrameTitleText:SetText(format(TRADE_SKILL_TITLE, GetTradeSkillLine()))
    TradeSkillSkillName:Hide()
    TradeSkillSkillIcon:Hide()
    TradeSkillRequirementLabel:Hide()
    TradeSkillRequirementText:SetText("")
    TradeSkillCollapseAllButton:Disable()
    for i=1, MAX_TRADE_SKILL_REAGENTS, 1 do
      getglobal("TradeSkillReagent"..i):Hide()
    end
  else
    TradeSkillSkillName:Show()
    TradeSkillSkillIcon:Show()
  end

  FauxScrollFrame_Update(TradeSkillListScrollFrame, numTradeSkills, TRADE_SKILLS_DISPLAYED, TRADE_SKILL_HEIGHT, nil, nil, nil, TradeSkillHighlightFrame, 293, 316 )

  TradeSkillHighlightFrame:Hide()
  for i=1, TRADE_SKILLS_DISPLAYED, 1 do
    local skillAt = i + skillOffset
    local skillIndex = state.trade_searched_recipes[skillAt] or 0
    local skillName, skillType, numAvailable, isExpanded = GetTradeSkillInfo(skillIndex)
    local skillButton = getglobal("TradeSkillSkill"..i)
    if ( skillAt <= numTradeSkills ) then
      if ( TradeSkillListScrollFrame:IsVisible() ) then
        skillButton:SetWidth(293)
      else
        skillButton:SetWidth(323)
      end
      local color = TradeSkillTypeColor[skillType]
      if ( color ) then
        skillButton:SetNormalFontObject(color.font)
      end

      skillButton:SetID(skillIndex)
      skillButton:Show()
      if ( skillType == "header" ) then
        skillButton:SetText(skillName)
        if ( isExpanded ) then
          skillButton:SetNormalTexture("Interface\\Buttons\\UI-MinusButton-Up")
        else
          skillButton:SetNormalTexture("Interface\\Buttons\\UI-PlusButton-Up")
        end
        getglobal("TradeSkillSkill"..i.."Highlight"):SetTexture("Interface\\Buttons\\UI-PlusButton-Hilight")
        getglobal("TradeSkillSkill"..i):UnlockHighlight()
      else
        if ( not skillName ) then
          return
        end
        skillButton:SetNormalTexture("")
        getglobal("TradeSkillSkill"..i.."Highlight"):SetTexture("")
        if ( numAvailable == 0 ) then
          skillButton:SetText(" "..skillName)
        else
          skillButton:SetText(" "..skillName.." ["..numAvailable.."]")
        end

        if ( GetTradeSkillSelectionIndex() == skillIndex ) then
          TradeSkillHighlightFrame:SetPoint("TOPLEFT", "TradeSkillSkill"..i, "TOPLEFT", 0, 0)
          TradeSkillHighlightFrame:Show()
          getglobal("TradeSkillSkill"..i):LockHighlight()
        else
          getglobal("TradeSkillSkill"..i):UnlockHighlight()
        end
      end

    else
      skillButton:Hide()
    end
  end

  local numHeaders = 0
  local notExpanded = 0
  for i=1, numTradeSkills, 1 do
    local skillName, skillType, numAvailable, isExpanded = GetTradeSkillInfo(i)
    if ( skillName and skillType == "header" ) then
      numHeaders = numHeaders + 1
      if ( not isExpanded ) then
        notExpanded = notExpanded + 1
      end
    end
    if ( GetTradeSkillSelectionIndex() == i ) then
      TradeSkillFrame.numAvailable = numAvailable
    end
  end
  if ( notExpanded ~= numHeaders ) then
    TradeSkillCollapseAllButton.collapsed = nil
    TradeSkillCollapseAllButton:SetNormalTexture("Interface\\Buttons\\UI-MinusButton-Up")
  else
    TradeSkillCollapseAllButton.collapsed = 1
    TradeSkillCollapseAllButton:SetNormalTexture("Interface\\Buttons\\UI-PlusButton-Up")
  end

  TradeSkillCollapseAllButton:Disable()
end

Package.SearchProfession = Module
